{
  Haar Wavelet v.0.5b

  (C)2K+1 by CARSTEN WAECHTER aka THE TOXIC AVENGER/AINC.

  This Unit/Prog is Public Domain..
   Feel free to use, enhance or even learn about this code..
  If you like it, please drop a message to:
   toxie@ainc.de
  To get the newest version of Haar Wavelet surf to:
   http://ainc.de

  All future changes will be listed below..
}

unit Wave;

interface

TYPE tbarray = array[0..100000000]of byte;
     tfarray = array[0..100000000]of double;
     tcarray = array[0..100000000]of cardinal;
     tiarray = array[0..100000000]of longint;
     twarray = array[0..100000000]of word;
     pbarray = ^tbarray;
     pfarray = ^tfarray;
     pcarray = ^tcarray;
     piarray = ^tiarray;
     pwarray = ^twarray;

//

     HWL_Header = packed record
                   ID           : array[0..2]of char; //'HWL'
                   width,height : word;     //X,Y
                   bpp          : byte;     //BitsPerPixel (HWL_GS,HWL_RGB)
                   depth        : byte;     //TransformationDepth
                   bitsr,bitsg,bitsb,datar,datag,datab : cardinal; //Bits/Data in Buffers
                   quantbits    : byte;     //Quantizer Bits Used (1..8)
                  end;
                  //Following the Header:
                  //1.   ((depth+2)*sizeof(double)) Bytes of Data for QuantizerFactors (One for HWL_GS, Three (R,G,B) for HWL_RGB)
                  //2.a. (bits*2/8) Bytes of Data for BitTable (HWL_ISO_ZERO,HWL_RLE_ZERO,etc)
                  //  b. (data*quantbits/8) Bytes of Data for DataTable
                  //  (One (R) for HWL_GS, Three (R,G,B) for HWL_RGB)

//

CONST HWL_GS=8;       //Valid BPP
      HWL_RGB=24;

      HWL_ISO_ZERO=0; //Valid En/Decoder-Values
      HWL_RLE_ZERO=1;
      HWL_POS=2;
      HWL_NEG=3;

//

PROCEDURE HWL2Mem(var pic : pointer; var head : HWL_Header; name : string; allocmem : boolean);
//For all who just want to get a HWL-File loaded into memory =)
//Everything else in this Unit is then TOTALLY unnecessary to work with / know about.
//Depending on the BPP the resulting "pic"-pointer will be filled with
//Bytes (HWL_GS) or Cardinals (HWL_RGB), the necessary infos are delivered inside "head".
//"allocmem" lets the procedure get memory for "pic" (TRUE) or awaits a already allocated "pic" (FALSE)

//

PROCEDURE WaveletGS(pic : pfarray; wl : pfarray; dx,dy,xres,yres,depth : cardinal);   //BMP to HWL
PROCEDURE DeWaveletGS(wl : pfarray; pic : pfarray; dx,dy,xres,yres,depth : cardinal); //Inverse

PROCEDURE WaveletQuantGS(wl : pfarray; wli : piarray; xres,yres,depth,bits : cardinal; quantfactor : pfarray); //Quantizer
PROCEDURE DeWaveletQuantGS(wli : piarray; wl : pfarray; xres,yres,depth : cardinal; quantfactor : pfarray);    //Inverse

FUNCTION  WaveletZeroOutGS(wl : pfarray; eps : double; bits : byte; xres,yres,depth : cardinal) : cardinal; //Kick Coefficients smaller "eps"
FUNCTION  WaveletCountZerosGS(wli : piarray; xres,yres,depth : cardinal) : cardinal; //Count overall Zeros in HWL

PROCEDURE ReadHWL(name : string; wlr,wlg,wlb : piarray; quantfactorr,quantfactorg,quantfactorb : pfarray); //Read HWL-File
PROCEDURE WriteHWL(name : string; wlr,wlg,wlb : piarray; xres,yres,bpp,depth : cardinal; quantfactorr,quantfactorg,quantfactorb : pfarray; quantbits : byte); //Write HWL-File

PROCEDURE ReadHeaderHWL(name : string; var head : HWL_Header); //Get HWL-File-Information

implementation


PROCEDURE HWL2Mem(var pic : pointer; var head : HWL_Header; name : string; allocmem : boolean);
 Var pic_qHWL,pic_qHWLr,pic_qHWLg,pic_qHWLb : piarray;
     pic_oHWL,pic_oHWLr,pic_oHWLg,pic_oHWLb : pfarray;
     pic_oBMP,pic_oBMPr,pic_oBMPg,pic_oBMPb : pfarray;
     pic_quantfactor,pic_quantfactorr,pic_quantfactorg,pic_quantfactorb : pfarray;

 PROCEDURE CopyDouble2ByteGS(src : PFarray; dest : PBarray); //GreyScale-FloatingPoint-BMP to Byte-BMP
  Var v : longint;
      col : byte;
 BEGIN
    for v:=0 to head.height*head.width-1 do begin
     if src^[v]>0.0 then begin
      if src^[v]<255.0 then begin
       col:=round(src^[v]);
      end else col:=$FF;
     end else col:=0;
     dest^[v]:=col;
    end;
 END;

 PROCEDURE CopyDouble2CardinalRGB(r,g,b : PFarray; dest : PCarray); //RGB-FloatingPoint-BMP to 32Bit-BMP
  Var v : longint;
      col : cardinal;
 BEGIN
    for v:=0 to head.height*head.width-1 do begin
     if b^[v]>0.0 then begin
      if b^[v]<255.0 then begin
       col:=round(b^[v]);
      end else col:=$FF;
     end else col:=0;

     if g^[v]>0.0 then begin
      if g^[v]<255.0 then begin
       col:=col or (round(g^[v])shl 8);
      end else col:=col or $FF00;
     end;

     if r^[v]>0.0 then begin
      if r^[v]<255.0 then begin
       col:=col or (round(r^[v])shl 16);
      end else col:=col or $FF0000;
     end;
     dest^[v]:=col;
    end;
 END;


BEGIN

  ReadHeaderHWL(name,head);

  if allocmem then begin
   if head.bpp=HWL_GS then reallocmem(pic,head.width*head.height)
    else if head.bpp=HWL_RGB then reallocmem(pic,head.width*head.height*4);
  end;

  if head.bpp=HWL_GS then begin
   getmem(pic_qHWL,head.width*head.height*sizeof(longint));
   getmem(pic_oHWL,head.width*head.height*sizeof(double));
   getmem(pic_oBMP,head.width*head.height*sizeof(double));
   getmem(pic_quantfactor,18*sizeof(double));

   ReadHWL(name,pic_qHWL,nil,nil,pic_quantfactor,nil,nil); //Load HWL

   DeWaveletQuantGS(pic_qHWL,pic_oHWL,head.width,head.height,head.depth,pic_quantfactor); //DeQuant

   DeWaveletGS(pic_oHWL,pic_oBMP,head.width,head.height,head.width,head.height,head.depth); //Restore BMP from HWL

   CopyDouble2ByteGS(pic_oBMP,pic);

   freemem(pic_quantfactor,18*sizeof(double));
   freemem(pic_oBMP,head.width*head.height*sizeof(double));
   freemem(pic_oHWL,head.width*head.height*sizeof(double));
   freemem(pic_qHWL,head.width*head.height*sizeof(longint));
  end else if head.bpp=HWL_RGB then begin
   getmem(pic_qHWLr,head.width*head.height*sizeof(longint));
   getmem(pic_oHWLr,head.width*head.height*sizeof(double));
   getmem(pic_oBMPr,head.width*head.height*sizeof(double));
   getmem(pic_qHWLg,head.width*head.height*sizeof(longint));
   getmem(pic_oHWLg,head.width*head.height*sizeof(double));
   getmem(pic_oBMPg,head.width*head.height*sizeof(double));
   getmem(pic_qHWLb,head.width*head.height*sizeof(longint));
   getmem(pic_oHWLb,head.width*head.height*sizeof(double));
   getmem(pic_oBMPb,head.width*head.height*sizeof(double));
   getmem(pic_quantfactorr,18*sizeof(double));
   getmem(pic_quantfactorg,18*sizeof(double));
   getmem(pic_quantfactorb,18*sizeof(double));

   ReadHWL(name,pic_qHWLr,pic_qHWLg,pic_qHWLb,pic_quantfactorr,pic_quantfactorg,pic_quantfactorb); //Load HWL

   DeWaveletQuantGS(pic_qHWLr,pic_oHWLr,head.width,head.height,head.depth,pic_quantfactorr); //DeQuant
   DeWaveletQuantGS(pic_qHWLg,pic_oHWLg,head.width,head.height,head.depth,pic_quantfactorg); //DeQuant
   DeWaveletQuantGS(pic_qHWLb,pic_oHWLb,head.width,head.height,head.depth,pic_quantfactorb); //DeQuant

   DeWaveletGS(pic_oHWLr,pic_oBMPr,head.width,head.height,head.width,head.height,head.depth); //Restore BMP from HWL
   DeWaveletGS(pic_oHWLg,pic_oBMPg,head.width,head.height,head.width,head.height,head.depth); //Restore BMP from HWL
   DeWaveletGS(pic_oHWLb,pic_oBMPb,head.width,head.height,head.width,head.height,head.depth); //Restore BMP from HWL

   CopyDouble2CardinalRGB(pic_oBMPr,pic_oBMPg,pic_oBMPb,pic);

   freemem(pic_quantfactorb,18*sizeof(double));
   freemem(pic_quantfactorg,18*sizeof(double));
   freemem(pic_quantfactorr,18*sizeof(double));
   freemem(pic_oBMPb,head.width*head.height*sizeof(double));
   freemem(pic_oHWLb,head.width*head.height*sizeof(double));
   freemem(pic_qHWLb,head.width*head.height*sizeof(longint));
   freemem(pic_oBMPg,head.width*head.height*sizeof(double));
   freemem(pic_oHWLg,head.width*head.height*sizeof(double));
   freemem(pic_qHWLg,head.width*head.height*sizeof(longint));
   freemem(pic_oBMPr,head.width*head.height*sizeof(double));
   freemem(pic_oHWLr,head.width*head.height*sizeof(double));
   freemem(pic_qHWLr,head.width*head.height*sizeof(longint));
  end;

END;

//

PROCEDURE WaveletGS(pic : pfarray; wl : pfarray; dx,dy,xres,yres,depth : cardinal);
 Var x,y : longint;
     tempx,tempy : pfarray;
     factor : double;
     offset : cardinal;
BEGIN
  factor:=(1.0/sqrt(2.0)); //Normalized Haar

  getmem(tempx,xres*yres*sizeof(double)); //Temporary Transformation-Storage
  getmem(tempy,xres*yres*sizeof(double));

  for y:=0 to dy-1 do begin               //Transform Rows
   offset:=y*xres;
   for x:=0 to (dx div 2)-1 do begin
    tempx^[x +offset]            := (pic^[x*2 +offset] + pic^[(x*2+1) +offset]) *factor; //LOW-PASS
    tempx^[(x+dx div 2) +offset] := (pic^[x*2 +offset] - pic^[(x*2+1) +offset]) *factor; //HIGH-PASS
   end;
  end;

  for x:=0 to dx-1 do                     //Transform Columns
   for y:=0 to (dy div 2)-1 do begin
    tempy^[x +y*xres]            := (tempx^[x +y*2*xres] + tempx^[x +(y*2+1)*xres]) *factor; //LOW-PASS
    tempy^[x +(y+dy div 2)*xres] := (tempx^[x +y*2*xres] - tempx^[x +(y*2+1)*xres]) *factor; //HIGH-PASS
   end;

  for y:=0 to dy-1 do
   move(tempy^[y*xres],wl^[y*xres],dx*sizeof(double)); //Copy to Wavelet

  freemem(tempx,xres*yres*sizeof(double)); //Free Temp-Storage
  freemem(tempy,xres*yres*sizeof(double));

  if depth>0 then waveletgs(wl,wl,dx div 2,dy div 2,xres,yres,depth-1); //Repeat for SubDivisionDepth
END;

PROCEDURE DeWaveletGS(wl : pfarray; pic : pfarray; dx,dy,xres,yres,depth : cardinal);
 Var x,y : longint;
     tempx,tempy : pfarray;
     offset : cardinal;
     factor : double;
BEGIN
  if depth>0 then dewaveletgs(wl,wl,dx div 2,dy div 2,xres,yres,depth-1); //Repeat for SubDivisionDepth

  factor:=(1.0/sqrt(2.0)); //Normalized Haar

  getmem(tempx,xres*yres*sizeof(double)); //Temporary Transformation-Storage
  getmem(tempy,xres*yres*sizeof(double));

  for x:=0 to dx-1 do                     //Inverse Transform Colums (fake: if (high-pass coefficient=0.0) and (surrounding high-pass coefficients=0.0) then interpolate between surrounding low-pass coefficients)
   for y:=0 to (dy div 2)-1 do
    if (wl^[x +(y+(dy div 2))*xres]<>0.0) or (y=(dy div 2)-1) or (y=0) or (dy<4) then begin //!UPDATED
     tempy^[x +y*2*xres]     :=(wl^[x +y*xres] + wl^[x +(y+(dy div 2))*xres])*factor; //LOW-PASS
     tempy^[x +(y*2+1)*xres] :=(wl^[x +y*xres] - wl^[x +(y+(dy div 2))*xres])*factor; //HIGH-PASS
    end else begin //!UPDATED
     if (wl^[x +(y-1+(dy div 2))*xres]=0.0) and (wl^[x +(y+1)*xres]<>wl^[x +y*xres]) and ((y=(dy div 2)-2) or (wl^[x +(y+1)*xres]<>wl^[x +(y+2)*xres])) then tempy^[x +y*2*xres]:=(wl^[x +y*xres]*0.8 + wl^[x +(y-1)*xres]*0.2)*factor //LOW-PASS
      else tempy^[x +y*2*xres]:=wl^[x +y*xres]*factor;
     if (wl^[x +(y+1+(dy div 2))*xres]=0.0) and (wl^[x +(y-1)*xres]<>wl^[x +y*xres]) and ((y=1) or (wl^[x +(y-1)*xres]<>wl^[x +(y-2)*xres])) then tempy^[x +(y*2+1)*xres]:=(wl^[x +y*xres]*0.8 + wl^[x +(y+1)*xres]*0.2)*factor //HIGH-PASS
      else tempy^[x +(y*2+1)*xres]:=wl^[x +y*xres]*factor;
    end;


  for y:=0 to dy-1 do begin               //Inverse Transform Rows (fake: if (high-pass coefficient=0.0) and (surrounding high-pass coefficients=0.0) then interpolate between surrounding low-pass coefficients)
   offset:=y*xres;
   for x:=0 to (dx div 2)-1 do
    if (tempy^[(x+dx div 2) +offset]<>0.0) or (x=(dx div 2)-1) or (x=0) or (dx<4) then begin //!UPDATED
     tempx^[x*2 +offset]     :=(tempy^[x +offset] + tempy^[(x+dx div 2) +offset])*factor; //LOW-PASS
     tempx^[(x*2+1) +offset] :=(tempy^[x +offset] - tempy^[(x+dx div 2) +offset])*factor; //HIGH-PASS
    end else begin //!UPDATED
     if (tempy^[(x-1+dx div 2) +offset]=0.0) and (tempy^[(x+1) +offset]<>tempy^[x +offset]) and ((x=(dx div 2)-2) or (tempy^[(x+1) +offset]<>tempy^[(x+2) +offset])) then tempx^[x*2 +offset]:=(tempy^[x +offset]*0.8 + tempy^[(x-1) +offset]*0.2)*factor //LOW-PASS
      else tempx^[x*2 +offset]:=tempy^[x +offset]*factor;
     if (tempy^[(x+1+dx div 2) +offset]=0.0) and (tempy^[(x-1) +offset]<>tempy^[x +offset]) and ((x=1) or (tempy^[(x-1) +offset]<>tempy^[(x-2) +offset])) then tempx^[(x*2+1) +offset]:=(tempy^[x +offset]*0.8 + tempy^[(x+1) +offset]*0.2)*factor //HIGH-PASS
      else tempx^[(x*2+1) +offset]:=tempy^[x +offset]*factor;
    end;
  end;


  for y:=0 to dy-1 do
   move(tempx^[y*xres],pic^[y*xres],dx*sizeof(double)); //Copy to Pic

  freemem(tempx,xres*yres*sizeof(double)); //Free Temp-Storage
  freemem(tempy,xres*yres*sizeof(double));
END;

PROCEDURE WaveletQuantGS(wl : pfarray; wli : piarray; xres,yres,depth,bits : cardinal; quantfactor : pfarray);
 Var startx,d,x,y,dx,dy,offset : longint;
     min,max,factor : double;
BEGIN
 //HIGH-PASS
 for d:=0 to depth do begin //Repeat for all SubDivisions
  min:=100000000.0;
  max:=-100000000.0;

  if d>0 then dy:=(yres shr d) -1 //Shifting Factors (to "navigate" within the Wavelet-HighPass)
   else dy:=yres-1;
  if d>0 then dx:=(xres shr d) -1
   else dx:=xres-1;

  for y:=0 to dy do begin //Get Minimum/Maximum Coefficient from current Subdivision-High-Pass
   offset:=y*xres;
   if (y>=(yres shr (d+1))) then startx:=0 //Only look inside the High-Pass
    else startx:=(xres shr (d+1));
   for x:=startx to dx do begin
    if wl^[x+offset]<min then min:=wl^[x+offset]; //Smaller/Bigger ??!
    if wl^[x+offset]>max then max:=wl^[x+offset];
   end;
  end;

  if (min<>0.0) or (max<>0.0) then //Calc Quantizer Factor
   if abs(min)>abs(max) then factor:=(1 shl bits -1)/abs(min)
   else factor:=(1 shl bits -1)/abs(max)
  else factor:=0.0;

  for y:=0 to dy do begin //Quantize (Linear Scale) the Coefficients
   offset:=y*xres;
   if (y>=(yres shr (d+1))) then startx:=0 //Only quantize inside the High-Pass
    else startx:=(xres shr (d+1));
   for x:=startx to dx do
    wli^[x+offset]:=round(wl^[x+offset]*factor);
  end;

  quantfactor^[d]:=factor; //Save current Quantize-Factor
 end;

  //LOW-PASS
  min:=100000000.0;
  max:=-100000000.0;

  for y:=0 to (yres shr (depth+1))-1 do begin //Get Minimum/Maximum Value from remaining Subdivision-Low-Pass
   offset:=y*xres;
   for x:=0 to (xres shr (depth+1))-1 do begin
    if wl^[x+offset]<min then min:=wl^[x+offset];
    if wl^[x+offset]>max then max:=wl^[x+offset];
   end;
  end;

  if (min<>0.0) or (max<>0.0) then begin //Calc Quantizer-factor
   if abs(min)>abs(max) then factor:=(1 shl bits -1)/abs(min)
    else factor:=(1 shl bits -1)/abs(max);
  end else factor:=0.0;

//  application.messagebox(@(floattostr(min)+'/'+floattostr(max)+'/'+floattostr(factor)+#0)[1],'min/max/factor',0); //TESTING PURPOSE

  for y:=0 to (yres shr (depth+1))-1 do begin //Quantize (Scale) Values
   offset:=y*xres;
   for x:=0 to (xres shr (depth+1))-1 do
    wli^[x+offset]:=round(wl^[x+offset]*factor);
  end;

  quantfactor^[depth+1]:=factor; //Save Factor

END;

PROCEDURE DeWaveletQuantGS(wli : piarray; wl : pfarray; xres,yres,depth : cardinal; quantfactor : pfarray);
 Var dx,dy,x,y,offset,d,startx : longint;
     factor : double;
BEGIN
 //HIGH-PASS
 for d:=0 to depth do begin        //Repeat for all SubDivisions
  if d>0 then dy:=(yres shr d) -1  //Shifting Factors (to "navigate" within the Wavelet-HighPass)
   else dy:=yres-1;
  if d>0 then dx:=(xres shr d) -1
   else dx:=xres-1;

  if quantfactor^[d]<>0.0 then factor:=1.0/quantfactor^[d] //Invert QuantFactor
   else factor:=0.0;

  for y:=0 to dy do begin
   offset:=y*xres;
   if (y>=(yres shr (d+1))) then startx:=0 //Only dequantize inside the High-Pass
    else startx:=(xres shr (d+1));
   for x:=startx to dx do
    wl^[x+offset]:=wli^[x+offset]*factor;  //dequant
  end;
 end;

  //LOW-PASS
  if quantfactor^[depth+1]<>0.0 then factor:=1.0/quantfactor^[depth+1] //Invert QuantFactor
   else factor:=0.0;

  for y:=0 to (yres shr (depth+1))-1 do begin
   offset:=y*xres;
   for x:=0 to (xres shr (depth+1))-1 do
    wl^[x+offset]:=wli^[x+offset]*factor; //Dequant LowPass
  end;
END;

FUNCTION  WaveletZeroOutGS(wl : pfarray; eps : double; bits : byte; xres,yres,depth : cardinal) : cardinal;
 Var x,y : longint;
     c,offset,startx : cardinal;
BEGIN
  c:=0; //Numbers of Coefficients kicked

  for y:=0 to yres-1 do begin
   offset:=y*xres;
   if (y>=(yres shr (depth+1))) then startx:=0  //Only High-Pass
    else startx:=(xres shr (depth+1));
   for x:=startx to xres-1 do                   //Search for all Coefficients<>0.0 and <=eps
    if (abs(wl^[x+offset])<=eps) and (wl^[x+offset]<>0.0) then begin wl^[x+offset]:=0.0; inc(c); end;
  end;

  WaveletZeroOutGS:=c;
END;

FUNCTION  WaveletCountZerosGS(wli : piarray; xres,yres,depth : cardinal) : cardinal;
 Var v : longint;
     c : cardinal;
BEGIN
  c:=0; //Count all Zeros in Wavelet

  for v:=0 to xres*yres-1 do
   if (wli^[v]=0) then inc(c);

  WaveletCountZerosGS:=c;
END;

PROCEDURE DeWaveletPackGS(wli : piarray; bitbuffer : pbarray; bitindex : cardinal; databuffer : pbarray; dataindex : cardinal; xres,yres,depth : cardinal);
 Var v : cardinal;
     bitcounter,datacounter : cardinal;
BEGIN
  v:=0;           //Index for the Wavelet
  bitcounter:=0;  //Index for the 2Bit-Table (HWL_ISO_ZERO,HWL_RLE_ZERO,HWL_POS,HWL_NEG)
  datacounter:=0; //Index for the 8Bit-DataTable

  while (bitcounter<=bitindex) and (datacounter<=dataindex) and (v<xres*yres) do begin
   if (bitbuffer^[bitcounter]=HWL_ISO_ZERO) then wli^[v]:=0            //Single Zero found
   else if bitbuffer^[bitcounter]=HWL_RLE_ZERO then begin              //(databuffer^[datacounter]+1) Zeros found
     fillchar(wli^[v],(databuffer^[datacounter]+1)*sizeof(longint),0);
     inc(v,databuffer^[datacounter]);
     inc(datacounter);
    end else if bitbuffer^[bitcounter]=HWL_POS then begin              //Positive Significant Coefficient found
     wli^[v]:=databuffer^[datacounter];
     inc(datacounter);
    end else begin                                                     //Negative Significant Coefficient found
     wli^[v]:=-databuffer^[datacounter];
     inc(datacounter);
    end;
   inc(bitcounter); //move on =)
   inc(v);
  end;
END;

PROCEDURE ReadHeaderHWL(name : string; var head : HWL_Header);
 Var f : file;
BEGIN
  Assignfile(f,name);
  filemode:=0;
  Reset(f,1);
  blockread(f,head,sizeof(head));
  Close(f);
END;

PROCEDURE ReadHWL(name : string; wlr,wlg,wlb : piarray; quantfactorr,quantfactorg,quantfactorb : pfarray);
 Var f : file;
     head : HWL_Header;
     bitbuffer,pbitbuffer : pbarray;
     databuffer,pdatabuffer : pbarray;
     bitindex,dataindex,pbits,pdata,data,data2 : cardinal;
     v : longint;
BEGIN
  Assignfile(f,name);
  filemode:=0;
  Reset(f,1);

  blockread(f,head,sizeof(head));

  //if head.ID<>'HWL' then ..

  if head.bpp=HWL_GS then begin //GREYSCALE
   blockread(f,quantfactorr^,(head.depth+2)*sizeof(double)); //Read Quantizer-Factors

   bitindex:=head.bitsr;  //2Bit-Packets in BitTable
   dataindex:=head.datar; //QuantBit-Packets in DataTable

   getmem(bitbuffer,bitindex);   //Alloc Mem for Byte-Storage of the Tables
   getmem(databuffer,dataindex);

   //UnPack the BitBuffer
   pbits:=bitindex div 4;

   if (bitindex mod 4=0) then begin
    getmem(pbitbuffer,pbits);    //Temp-Storage for packed BitTable
    blockread(f,pbitbuffer^,pbits);
   end else begin
    getmem(pbitbuffer,pbits+1);
    blockread(f,pbitbuffer^,pbits+1);
   end;

   for v:=0 to pbits-1 do begin //4 2Bit-Pakets match one Byte =)
    bitbuffer^[v*4]:=pbitbuffer^[v] and 3;
    bitbuffer^[v*4+1]:=(pbitbuffer^[v] and (3 shl 2)) shr 2;
    bitbuffer^[v*4+2]:=(pbitbuffer^[v] and (3 shl 4)) shr 4;
    bitbuffer^[v*4+3]:=(pbitbuffer^[v] and (3 shl 6)) shr 6;
   end;

   if (bitindex mod 4>0) then begin
    bitbuffer^[pbits*4]:=pbitbuffer^[pbits] and 3;
    if (bitindex mod 4>1) then bitbuffer^[pbits*4+1]:=(pbitbuffer^[pbits] and (3 shl 2)) shr 2;
    if (bitindex mod 4>2) then bitbuffer^[pbits*4+2]:=(pbitbuffer^[pbits] and (3 shl 4)) shr 4;
   end;

   if (bitindex mod 4=0) then freemem(pbitbuffer,pbits)
    else freemem(pbitbuffer,pbits+1);
   //End UnPack

   //Unpack the DataBuffer
   if ((dataindex*head.quantbits) mod 8=0) then pdata:=(dataindex*head.quantbits) div 8
    else pdata:=(dataindex*head.quantbits) div 8 +1;

   getmem(pdatabuffer,pdata);       //Temp-Storage for packed DataTable
   blockread(f,pdatabuffer^,pdata);

   pbits:=0;
   for v:=0 to dataindex-1 do begin //ReStore Bytes out of the "packed" BitStream
    if (pbits mod 8=0) then begin
     data:=pdatabuffer^[pbits div 8] and ((1 shl head.quantbits)-1);
    end else begin
     data:=(cardinal(pdatabuffer^[pbits div 8]) shr (pbits mod 8)) and ((1 shl head.quantbits)-1);
     if ((pbits+head.quantbits) div 8>pbits div 8) then begin
      data2:=cardinal(pdatabuffer^[(pbits div 8)+1]) and ((1 shl ((pbits+head.quantbits) mod 8))-1);
      data:=data or (data2 shl (8-(pbits mod 8)));
     end;
    end;
    databuffer^[v]:=data;
    inc(pbits,head.quantbits);
   end;

   freemem(pdatabuffer,pdata);
   //End Unpack

   DeWaveletPackGS(wlr,bitbuffer,bitindex,databuffer,dataindex,head.width,head.height,head.depth); //Restore Wavelet from the Byte-Tables

   freemem(bitbuffer,bitindex);
   freemem(databuffer,dataindex);
  end else if head.bpp=HWL_RGB then begin //RGB
   blockread(f,quantfactorr^,(head.depth+2)*sizeof(double)); //Read Quantizer-Factors
   blockread(f,quantfactorg^,(head.depth+2)*sizeof(double));
   blockread(f,quantfactorb^,(head.depth+2)*sizeof(double));

   //RED

   bitindex:=head.bitsr;  //2Bit-Packets in BitTable (Red)
   dataindex:=head.datar; //QuantBit-Packets in DataTable

   getmem(bitbuffer,bitindex);  //Alloc Mem for Byte-Storage of the Tables
   getmem(databuffer,dataindex);

   //UnPack the BitBuffer
   pbits:=bitindex div 4;

   if (bitindex mod 4=0) then begin
    getmem(pbitbuffer,pbits);    //Temp-Storage for packed BitTable
    blockread(f,pbitbuffer^,pbits);
   end else begin
    getmem(pbitbuffer,pbits+1);
    blockread(f,pbitbuffer^,pbits+1);
   end;

   for v:=0 to pbits-1 do begin //4 2Bit-Pakets match one Byte =)
    bitbuffer^[v*4]:=pbitbuffer^[v] and 3;
    bitbuffer^[v*4+1]:=(pbitbuffer^[v] and (3 shl 2)) shr 2;
    bitbuffer^[v*4+2]:=(pbitbuffer^[v] and (3 shl 4)) shr 4;
    bitbuffer^[v*4+3]:=(pbitbuffer^[v] and (3 shl 6)) shr 6;
   end;

   if (bitindex mod 4>0) then begin
    bitbuffer^[pbits*4]:=pbitbuffer^[pbits] and 3;
    if (bitindex mod 4>1) then bitbuffer^[pbits*4+1]:=(pbitbuffer^[pbits] and (3 shl 2)) shr 2;
    if (bitindex mod 4>2) then bitbuffer^[pbits*4+2]:=(pbitbuffer^[pbits] and (3 shl 4)) shr 4;
   end;

   if (bitindex mod 4=0) then freemem(pbitbuffer,pbits)
    else freemem(pbitbuffer,pbits+1);
   //End UnPack

   //Unpack the DataBuffer
   if ((dataindex*head.quantbits) mod 8=0) then pdata:=(dataindex*head.quantbits) div 8
    else pdata:=(dataindex*head.quantbits) div 8 +1;

   getmem(pdatabuffer,pdata); //Temp-Storage for packed DataTable
   blockread(f,pdatabuffer^,pdata);

   pbits:=0;
   for v:=0 to dataindex-1 do begin //ReStore Bytes out of the "packed" BitStream
    if (pbits mod 8=0) then begin
     data:=pdatabuffer^[pbits div 8] and ((1 shl head.quantbits)-1);
    end else begin
     data:=(cardinal(pdatabuffer^[pbits div 8]) shr (pbits mod 8)) and ((1 shl head.quantbits)-1);
     if ((pbits+head.quantbits) div 8>pbits div 8) then begin
      data2:=cardinal(pdatabuffer^[(pbits div 8)+1]) and ((1 shl ((pbits+head.quantbits) mod 8))-1);
      data:=data or (data2 shl (8-(pbits mod 8)));
     end;
    end;
    databuffer^[v]:=data;
    inc(pbits,head.quantbits);
   end;

   freemem(pdatabuffer,pdata);
   //End Unpack

   DeWaveletPackGS(wlr,bitbuffer,bitindex,databuffer,dataindex,head.width,head.height,head.depth);

   freemem(bitbuffer,bitindex);
   freemem(databuffer,dataindex);

   //GREEN

   bitindex:=head.bitsg;  //2Bit-Packets in BitTable (Green)
   dataindex:=head.datag; //QuantBit-Packets in DataTable

   getmem(bitbuffer,bitindex);  //Alloc Mem for Byte-Storage of the Tables
   getmem(databuffer,dataindex);

   //UnPack the BitBuffer
   pbits:=bitindex div 4;

   if (bitindex mod 4=0) then begin
    getmem(pbitbuffer,pbits);    //Temp-Storage for packed BitTable
    blockread(f,pbitbuffer^,pbits);
   end else begin
    getmem(pbitbuffer,pbits+1);
    blockread(f,pbitbuffer^,pbits+1);
   end;

   for v:=0 to pbits-1 do begin //4 2Bit-Pakets match one Byte =)
    bitbuffer^[v*4]:=pbitbuffer^[v] and 3;
    bitbuffer^[v*4+1]:=(pbitbuffer^[v] and (3 shl 2)) shr 2;
    bitbuffer^[v*4+2]:=(pbitbuffer^[v] and (3 shl 4)) shr 4;
    bitbuffer^[v*4+3]:=(pbitbuffer^[v] and (3 shl 6)) shr 6;
   end;

   if (bitindex mod 4>0) then begin
    bitbuffer^[pbits*4]:=pbitbuffer^[pbits] and 3;
    if (bitindex mod 4>1) then bitbuffer^[pbits*4+1]:=(pbitbuffer^[pbits] and (3 shl 2)) shr 2;
    if (bitindex mod 4>2) then bitbuffer^[pbits*4+2]:=(pbitbuffer^[pbits] and (3 shl 4)) shr 4;
   end;

   if (bitindex mod 4=0) then freemem(pbitbuffer,pbits)
    else freemem(pbitbuffer,pbits+1);
   //End UnPack

   //Unpack the DataBuffer
   if ((dataindex*head.quantbits) mod 8=0) then pdata:=(dataindex*head.quantbits) div 8
    else pdata:=(dataindex*head.quantbits) div 8 +1;

   getmem(pdatabuffer,pdata); //Temp-Storage for packed DataTable
   blockread(f,pdatabuffer^,pdata);

   pbits:=0;
   for v:=0 to dataindex-1 do begin //ReStore Bytes out of the "packed" BitStream
    if (pbits mod 8=0) then begin
     data:=pdatabuffer^[pbits div 8] and ((1 shl head.quantbits)-1);
    end else begin
     data:=(cardinal(pdatabuffer^[pbits div 8]) shr (pbits mod 8)) and ((1 shl head.quantbits)-1);
     if ((pbits+head.quantbits) div 8>pbits div 8) then begin
      data2:=cardinal(pdatabuffer^[(pbits div 8)+1]) and ((1 shl ((pbits+head.quantbits) mod 8))-1);
      data:=data or (data2 shl (8-(pbits mod 8)));
     end;
    end;
    databuffer^[v]:=data;
    inc(pbits,head.quantbits);
   end;

   freemem(pdatabuffer,pdata);
   //End Unpack

   DeWaveletPackGS(wlg,bitbuffer,bitindex,databuffer,dataindex,head.width,head.height,head.depth);

   freemem(bitbuffer,bitindex);
   freemem(databuffer,dataindex);

   //BLUE

   bitindex:=head.bitsb;  //2Bit-Packets in BitTable (Blue)
   dataindex:=head.datab; //QuantBit-Packets in DataTable

   getmem(bitbuffer,bitindex);   //Alloc Mem for Byte-Storage of the Tables
   getmem(databuffer,dataindex);

   //UnPack the BitBuffer
   pbits:=bitindex div 4;

   if (bitindex mod 4=0) then begin
    getmem(pbitbuffer,pbits);    //Temp-Storage for packed BitTable
    blockread(f,pbitbuffer^,pbits);
   end else begin
    getmem(pbitbuffer,pbits+1);
    blockread(f,pbitbuffer^,pbits+1);
   end;

   for v:=0 to pbits-1 do begin //4 2Bit-Pakets match one Byte =)
    bitbuffer^[v*4]:=pbitbuffer^[v] and 3;
    bitbuffer^[v*4+1]:=(pbitbuffer^[v] and (3 shl 2)) shr 2;
    bitbuffer^[v*4+2]:=(pbitbuffer^[v] and (3 shl 4)) shr 4;
    bitbuffer^[v*4+3]:=(pbitbuffer^[v] and (3 shl 6)) shr 6;
   end;

   if (bitindex mod 4>0) then begin
    bitbuffer^[pbits*4]:=pbitbuffer^[pbits] and 3;
    if (bitindex mod 4>1) then bitbuffer^[pbits*4+1]:=(pbitbuffer^[pbits] and (3 shl 2)) shr 2;
    if (bitindex mod 4>2) then bitbuffer^[pbits*4+2]:=(pbitbuffer^[pbits] and (3 shl 4)) shr 4;
   end;

   if (bitindex mod 4=0) then freemem(pbitbuffer,pbits)
    else freemem(pbitbuffer,pbits+1);
   //End UnPack

   //Unpack the DataBuffer
   if ((dataindex*head.quantbits) mod 8=0) then pdata:=(dataindex*head.quantbits) div 8
    else pdata:=(dataindex*head.quantbits) div 8 +1;

   getmem(pdatabuffer,pdata); //Temp-Storage for packed DataTable
   blockread(f,pdatabuffer^,pdata);

   pbits:=0;
   for v:=0 to dataindex-1 do begin //ReStore Bytes out of the "packed" BitStream
    if (pbits mod 8=0) then begin
     data:=pdatabuffer^[pbits div 8] and ((1 shl head.quantbits)-1);
    end else begin
     data:=(cardinal(pdatabuffer^[pbits div 8]) shr (pbits mod 8)) and ((1 shl head.quantbits)-1);
     if ((pbits+head.quantbits) div 8>pbits div 8) then begin
      data2:=cardinal(pdatabuffer^[(pbits div 8)+1]) and ((1 shl ((pbits+head.quantbits) mod 8))-1);
      data:=data or (data2 shl (8-(pbits mod 8)));
     end;
    end;
    databuffer^[v]:=data;
    inc(pbits,head.quantbits);
   end;

   freemem(pdatabuffer,pdata);
   //End Unpack

   DeWaveletPackGS(wlb,bitbuffer,bitindex,databuffer,dataindex,head.width,head.height,head.depth);

   freemem(bitbuffer,bitindex);
   freemem(databuffer,dataindex);
  end;

  Closefile(f);
END;

PROCEDURE WaveletPackGS(wl : piarray; bitbuffer : pbarray; var bitindex : cardinal; databuffer : pbarray; var dataindex : cardinal; xres,yres,depth,bits : cardinal);
Var v,c,maxlength,minlength : cardinal;

 FUNCTION Zeros2Come : cardinal; //Count overall Zeros AFTER the current position in the Wavelet
  Var t : cardinal;
 BEGIN
   t:=1;
   while (v+t<xres*yres-1) and (wl^[v+t]=0) and (t<maxlength) do inc(t);
   zeros2come:=t-1;
 END;

BEGIN
   maxlength:=1 shl bits;     //Maximum RLE-length (depends on the QuantizerBits!)
   minlength:=(bits+1) div 2; //Minimum RLE-length (otherwise the File gets bigger then possible!)

   v:=0;
   while (v<xres*yres) do begin //Go through Wavelet
     if wl^[v]=0 then begin     //Found a Zero
      c:=zeros2come;            //Count Zeros afterwards
      if c<=minlength then bitbuffer^[bitindex]:=HWL_ISO_ZERO //One (or only a few) Zeros found -> Encode Single Zero
       else begin
        bitbuffer^[bitindex]:=HWL_RLE_ZERO; //Found a lotta Zeros =)
        databuffer^[dataindex]:=c;          //Store the amount in the DataBuffer
        inc(dataindex);                     //inc databuffer-index
        inc(v,c);                           //inc position in wavelet
       end;
     end else if wl^[v]>0 then begin   //Significant Positive
      bitbuffer^[bitindex]:=HWL_POS;
      databuffer^[dataindex]:=wl^[v];  //store in databuffer
      inc(dataindex);
     end else begin                    //Significant Negative
      bitbuffer^[bitindex]:=HWL_NEG;
      databuffer^[dataindex]:=-wl^[v]; //store in databuffer
      inc(dataindex);
     end;
    inc(bitindex);                     //continue in wavelet/bitTable
    inc(v);
   end;
END;

PROCEDURE WriteHWL(name : string; wlr,wlg,wlb : piarray; xres,yres,bpp,depth : cardinal; quantfactorr,quantfactorg,quantfactorb : pfarray; quantbits : byte);
 Var f : file;
     head : HWL_Header;
     bitbuffer : pbarray;
     databuffer : pbarray;
     bitindex,dataindex,pbits,pbits2,data,data2 : cardinal;
     v : longint;
BEGIN
  head.ID:='HWL';    //Prepare HWL-Header
  head.width:=xres;
  head.height:=yres;
  head.bpp:=bpp;
  head.depth:=depth;
  head.quantbits:=quantbits;

  getmem(bitbuffer,xres*yres);  //Alloc Mem for the Bit/Data-Table
  getmem(databuffer,xres*yres);

  Assignfile(f,name);
  filemode:=2;
  Rewrite(f,1);

  if bpp=HWL_GS then begin //Greyscale
   bitindex:=0;
   dataindex:=0;

   WaveletPackGS(wlr,bitbuffer,bitindex,databuffer,dataindex,xres,yres,depth,quantbits); //Pack the Wavelet into Byte-Tables

   //Pack the BitBuffer
   pbits:=bitindex div 4;

   for v:=0 to pbits-1 do //4 2Bit-Pakets match one Byte =)
    bitbuffer^[v]:=bitbuffer^[v*4] or (bitbuffer^[v*4+1] shl 2) or (bitbuffer^[v*4+2] shl 4) or (bitbuffer^[v*4+3] shl 6);

   case (bitindex mod 4) of
    1: bitbuffer^[pbits]:=bitbuffer^[pbits*4];
    2: bitbuffer^[pbits]:=bitbuffer^[pbits*4] or (bitbuffer^[pbits*4+1] shl 2);
    3: bitbuffer^[pbits]:=bitbuffer^[pbits*4] or (bitbuffer^[pbits*4+1] shl 2) or (bitbuffer^[pbits*4+2] shl 4);
   end;
   //End Pack

   head.bitsr:=bitindex; //Output Head,QuantizerFactors and BitTable
   head.datar:=dataindex;
   blockwrite(f,head,sizeof(head));
   blockwrite(f,quantfactorr^,(depth+2)*sizeof(double));
   if (bitindex mod 4=0) then blockwrite(f,bitbuffer^,pbits)
    else blockwrite(f,bitbuffer^,pbits+1);

   //Pack the DataBuffer
   pbits:=0;
   for v:=0 to dataindex-1 do begin //Pack single Bytes into a BitStream (depending on QuantizerBits)
    if (pbits mod 8=0) then begin
     data:=databuffer^[v];
    end else begin
     data:=cardinal(databuffer^[v]) shl (pbits mod 8);
     if ((pbits+quantbits) div 8>pbits div 8) then begin
      data2:=data shr 8;
      data:=data and $FF;
      databuffer^[(pbits div 8)+1]:=data2;
     end;
     data:=data or (cardinal(databuffer^[pbits div 8]) and ((1 shl (pbits mod 8))-1));
    end;
    databuffer^[pbits div 8]:=data;
    inc(pbits,quantbits);
   end;
   //End Pack

   if (pbits mod 8=0) then blockwrite(f,databuffer^,pbits div 8) //Output Data-Bit-Table
    else blockwrite(f,databuffer^,(pbits div 8)+1);
  end else if bpp=HWL_RGB then begin //RGB
   //RED
   bitindex:=0;
   dataindex:=0;

   WaveletPackGS(wlr,bitbuffer,bitindex,databuffer,dataindex,xres,yres,depth,quantbits); //Pack the Wavelet into Byte-Tables

   //Pack the BitBuffer
   pbits:=bitindex div 4;

   for v:=0 to pbits-1 do //4 2Bit-Pakets match one Byte =)
    bitbuffer^[v]:=bitbuffer^[v*4] or (bitbuffer^[v*4+1] shl 2) or (bitbuffer^[v*4+2] shl 4) or (bitbuffer^[v*4+3] shl 6);

   case (bitindex mod 4) of
    1: bitbuffer^[pbits]:=bitbuffer^[pbits*4];
    2: bitbuffer^[pbits]:=bitbuffer^[pbits*4] or (bitbuffer^[pbits*4+1] shl 2);
    3: bitbuffer^[pbits]:=bitbuffer^[pbits*4] or (bitbuffer^[pbits*4+1] shl 2) or (bitbuffer^[pbits*4+2] shl 4);
   end;
   //End Pack

   if (bitindex mod 4=0) then pbits2:=pbits
    else pbits2:=pbits+1;

   //Pack the DataBuffer
   pbits:=0;
   for v:=0 to dataindex-1 do begin //Pack single Bytes into a BitStream (depending on QuantizerBits)
    if (pbits mod 8=0) then begin
     data:=databuffer^[v];
    end else begin
     data:=cardinal(databuffer^[v]) shl (pbits mod 8);
     if ((pbits+quantbits) div 8>pbits div 8) then begin
      data2:=data shr 8;
      data:=data and $FF;
      databuffer^[(pbits div 8)+1]:=data2;
     end;
     data:=data or (cardinal(databuffer^[pbits div 8]) and ((1 shl (pbits mod 8))-1));
    end;
    databuffer^[pbits div 8]:=data;
    inc(pbits,quantbits);
   end;
   //End Pack

   head.bitsr:=bitindex;  //Output Head,QuantizerFactors (Red,Green,Blue), BitTable (Red) and DataTable (Red)
   head.datar:=dataindex;
   blockwrite(f,head,sizeof(head));
   blockwrite(f,quantfactorr^,(depth+2)*sizeof(double));
   blockwrite(f,quantfactorg^,(depth+2)*sizeof(double));
   blockwrite(f,quantfactorb^,(depth+2)*sizeof(double));
   blockwrite(f,bitbuffer^,pbits2);
   if (pbits mod 8=0) then blockwrite(f,databuffer^,pbits div 8)
    else blockwrite(f,databuffer^,(pbits div 8)+1);

   //GREEN

   bitindex:=0;
   dataindex:=0;

   WaveletPackGS(wlg,bitbuffer,bitindex,databuffer,dataindex,xres,yres,depth,quantbits); //Pack the Wavelet into Byte-Tables

   //Pack the BitBuffer
   pbits:=bitindex div 4;

   for v:=0 to pbits-1 do //4 2Bit-Pakets match one Byte =)
    bitbuffer^[v]:=bitbuffer^[v*4] or (bitbuffer^[v*4+1] shl 2) or (bitbuffer^[v*4+2] shl 4) or (bitbuffer^[v*4+3] shl 6);

   case (bitindex mod 4) of
    1: bitbuffer^[pbits]:=bitbuffer^[pbits*4];
    2: bitbuffer^[pbits]:=bitbuffer^[pbits*4] or (bitbuffer^[pbits*4+1] shl 2);
    3: bitbuffer^[pbits]:=bitbuffer^[pbits*4] or (bitbuffer^[pbits*4+1] shl 2) or (bitbuffer^[pbits*4+2] shl 4);
   end;
   //End Pack

   if (bitindex mod 4=0) then pbits2:=pbits
    else pbits2:=pbits+1;

   //Pack the DataBuffer
   pbits:=0;
   for v:=0 to dataindex-1 do begin //Pack single Bytes into a BitStream (depending on QuantizerBits)
    if (pbits mod 8=0) then begin
     data:=databuffer^[v];
    end else begin
     data:=cardinal(databuffer^[v]) shl (pbits mod 8);
     if ((pbits+quantbits) div 8>pbits div 8) then begin
      data2:=data shr 8;
      data:=data and $FF;
      databuffer^[(pbits div 8)+1]:=data2;
     end;
     data:=data or (cardinal(databuffer^[pbits div 8]) and ((1 shl (pbits mod 8))-1));
    end;
    databuffer^[pbits div 8]:=data;
    inc(pbits,quantbits);
   end;
   //End Pack

   head.bitsg:=bitindex; //Output BitTable (Green) and DataTable (Green)
   head.datag:=dataindex;
   blockwrite(f,bitbuffer^,pbits2);
   if (pbits mod 8=0) then blockwrite(f,databuffer^,pbits div 8)
    else blockwrite(f,databuffer^,(pbits div 8)+1);

   //BLUE

   bitindex:=0;
   dataindex:=0;

   WaveletPackGS(wlb,bitbuffer,bitindex,databuffer,dataindex,xres,yres,depth,quantbits); //Pack the Wavelet into Byte-Tables

   //Pack the BitBuffer
   pbits:=bitindex div 4;

   for v:=0 to pbits-1 do //4 2Bit-Pakets match one Byte =)
    bitbuffer^[v]:=bitbuffer^[v*4] or (bitbuffer^[v*4+1] shl 2) or (bitbuffer^[v*4+2] shl 4) or (bitbuffer^[v*4+3] shl 6);

   case (bitindex mod 4) of
    1: bitbuffer^[pbits]:=bitbuffer^[pbits*4];
    2: bitbuffer^[pbits]:=bitbuffer^[pbits*4] or (bitbuffer^[pbits*4+1] shl 2);
    3: bitbuffer^[pbits]:=bitbuffer^[pbits*4] or (bitbuffer^[pbits*4+1] shl 2) or (bitbuffer^[pbits*4+2] shl 4);
   end;
   //End Pack

   if (bitindex mod 4=0) then pbits2:=pbits
    else pbits2:=pbits+1;

   //Pack the DataBuffer
   pbits:=0;
   for v:=0 to dataindex-1 do begin //Pack single Bytes into a BitStream (depending on QuantizerBits)
    if (pbits mod 8=0) then begin
     data:=databuffer^[v];
    end else begin
     data:=cardinal(databuffer^[v]) shl (pbits mod 8);
     if ((pbits+quantbits) div 8>pbits div 8) then begin
      data2:=data shr 8;
      data:=data and $FF;
      databuffer^[(pbits div 8)+1]:=data2;
     end;
     data:=data or (cardinal(databuffer^[pbits div 8]) and ((1 shl (pbits mod 8))-1));
    end;
    databuffer^[pbits div 8]:=data;
    inc(pbits,quantbits);
   end;
   //End Pack

   head.bitsb:=bitindex; //Output BitTable (Blue) and DataTable (Blue)
   head.datab:=dataindex;
   blockwrite(f,bitbuffer^,pbits2);
   if (pbits mod 8=0) then blockwrite(f,databuffer^,pbits div 8)
    else blockwrite(f,databuffer^,(pbits div 8)+1);

   seek(f,0);
   blockwrite(f,head,sizeof(head)); //Output Head (again =)
  end;

  freemem(bitbuffer,xres*yres);
  freemem(databuffer,xres*yres);

  Closefile(f);

  {application.messagebox(@(inttostr(pbits)+#0)[1],'bits',0);
  application.messagebox(@(inttostr(dataindex)+#0)[1],'data',0);//TEST PURPOSE}
END;

end.

{HISTORY: 02.02.2001: -added a little 'trick' to the DeWaveletGS-Routine..
                       it's based upon the idea of more complex wavelets, which
                       try to filter more "information" (-> pixeldata) into
                       single low/high-pass coefficients..
                       the 'trick' in my routine uses an idea i adopted from
                       alan watt's article about wavelets in his "3d games vol.1"-book..
                       he uses linear interpolation of the transformed coefficients
                       (-> results in an effect like gouraud shading =)
                       together with a (progressive) quadtree representation of the coefficient data..
                       DeWaveletGS simulates (read: "fakes" ;) this effect by interpolating
                       between the low-pass coefficients (if the related high-pass coefficient=0.0 (-> kicked data))..

                       Result: Amazing Quality Improvement! (-> CHECK IT OUT! 8-)
                       Best Thing: All previously stored .hwl-files remain valid.. 

                       (Search for //!UPDATED in the source)

          04.02.2001: -little bugfix for high contrasts (-> resulted in 'false' interpolation producing ugly artefacts!)
}